"""
nilutility/statistics.py
$Id: statistics.py 2556 2020-12-02 20:12:20Z pe $

Basic statistics routines.
"""

import math
from nilutility.ndecimal import Decimal
from bisect import insort, bisect_left
from collections import deque, defaultdict

def median(lst, is_sorted=False, contains_none=True):
    """
    Calculates the median of a list of values.
    Parameters:
        lst         list of values
        is_sorted   bool, speed up the calculation if the list is already sorted
        contains_none
                    bool, speed up the calculation if the list does not contain
                    Nones
    Returns:
        median value
    """
    if contains_none:
        lst = [x for x in lst if x is not None]
    if not is_sorted:
        lst = sorted(lst)
    # do not check is_sorted or contains_none:
    #  this would be a performace killer (ca. factor 1000 for contains_None)
    #  is_sorted would be even more expansive to check
    return _median_fast(lst)

def _median_fast(lst):
    """
    A raw and fast version of calculating the median if one can guarantee that
    the list is sorted and there aer no None's in the list.
    !!! Fast version, list must be sorted and may not contain None values!
    For unsorted lists or lists containing None's use median()
    Parameters:
        lst   list of values
    Returns:
        median of values, None in case of empty list
    """
    len_ = len(lst)
    if len_ == 0:
        # empty list, no median
        return None
    if len_ % 2:
        return lst[len_//2]
    try:
        ret = (lst[len_//2-1] + lst[len_//2]) / 2.0
    except TypeError:
        # in case where /2.0 is not allowed (e.g. datetime.timedelta)
        ret = (lst[len_//2-1] + lst[len_//2]) / 2
    return ret

def modified_median(lst, **kwargs):
    """
    This is a modified median. If less then 75% of th evalues are the same
    but still the most abundant 2 different values make at least 90%:
    use the arithmetic mean of those two elements.
    In all other cases, the median is returned.
    
    Parameters:
        lst        list of values
        **kwargs   keyword arguments to be passed to the fallback median method

    Used to find a representative value for the time resolution of time series.
    Before, we used the median (if majority of the samples are 1h, use 1h even
    though there are some outliers...)
    This modified median tackles the followong edge case:
    Samples taken at 3+4 days interval. The median returns 3 if the series
    starts with 3d and ends with 3 h. It returns 4d if the sequence starts with
    4d and ends with 4d. And it resturn 3.5d (84h) if there is the same number
    of both.
    """
    if len(lst) > 3:
        histo = defaultdict(int)
        for elem in lst:
            histo[elem] += 1
        res = sorted(histo.items(), key=lambda x: x[1], reverse=True)
        # how many of the most abundant value?
        if res[0][1] >= len(lst)*0.75:
            # >= 75% og all values are the same
            return res[0][0]
        if res[0][1] + res[1][1] >= len(lst)*0.9:
            # >= 90% of all values are the most abundant 2)
            return (res[0][0] + res[1][0]) / 2
    # else return median
    return median(lst, **kwargs)

def running_median(lst, radius):
    """
    Running median. The running median for an element n is the median of element
    n and radius elements before and after (Window size = 2 * radius + 1).
    For the first (and last) n elements, the number of elements before (after)
    is iof course shorter then n.

    Parameters:
        lst    list of values
        radius the radius (how many elements left and right on an element n
               should be included).
    Returns:
        list with running medians.

    Highly performant implementation, especially with bigger window sizes about
    100-1000 times faster than more naive approaches.
    Makes use of the fact that only 2 elements are changed within the window for
    each iteration.

    Performance of different approches has been tested in a small test setup:
    https://git.nilu.no/ebas/scripts/test/blob/master/running_median_test.py
    The bigger the used window sizes, the clearer the advantages of this
    implementation is visible.

    A double-ended queue (deque; with fast append and pop on both sides) is used
    to keep track of the current window. In parallel a list of the window
    elements is always kept sorted, using the highly performant bisect methods
    (insort, bisect_left) to maintain the sorted list.
    """
    result = []
    # Initial window: n elements right of the value plus the value itself (=n+1
    # values).
    # We preload the window with n values. in the first iteration + 1 value
    # is added
    deq_win = deque(lst[0:radius])  # The window as double ended queue
    # The window as sorted list:
    srt_win = sorted([x for x in deq_win if x is not None])
    for item in lst[radius:]:
        # Iterate through elements from radius to the end
        # Item is the rightmost element to be appended to the window:
        deq_win.append(item)
        if item is not None:
            insort(srt_win, item)
        # Delete one element if the window size is exceeded:
        if len(deq_win) > 2 * radius + 1:
            tbd = deq_win.popleft()  # item to be deleted
            if tbd is not None:
                del srt_win[bisect_left(srt_win, tbd)]
        # Calculate and append the median value:
        # (we know the window is sorted and contains no None values, use
        # _median_fast directly)
        result.append(_median_fast(srt_win))
    # Process the final elements (no new values, but pop values from the
    # left of the window.
    # Usually this is radius iterations, but specail case when list size <
    # rarius.
    for i in range(len(lst) - len(result)):
        # with list much longer then radius, we would just unconditionally pop
        # one element per loop. But when list length < 2*r+1 (i.e. full window
        # size is never reached), we pop elements too early for the last r
        # elements. Thus we depend on i:.
        # Example: [1., 2., 3., 4.], radius = 2
        if len(deq_win) > 2 * radius - i:
            tbd = deq_win.popleft()  # item to be deleted
            if tbd is not None:
                del srt_win[bisect_left(srt_win, tbd)]
        # Calculate and append the median value:
        # (we know the window is sorted and contains no None values, use
        # _median_fast directly)
        result.append(_median_fast(srt_win))
    return result

def stddev(lst, contains_none=True):
    """
    Calculates the stddev of a list of values.
    stddev of a sample of a population (i.e. df = n-1).
    Parameters:
        lst         list of values
        contains_none
                    bool, speed up the calculation if the list does not contain
                    Nones
    Returns:
        stddev value
    """
    if contains_none:
        lst = [x for x in lst if x is not None]
    return _stddev_fast(lst)

def _stddev_fast(lst):
    """
    A raw and fast version of calculating the stddev of a list of values (stddev
    of a sample of a population (i.e. df = n-1)).
    !!! Fast version, list must not contain None values!
    For lists containing None's use stddev()
    Parameters:
        lst   list of values (may not contain None's)
    Returns:
        stddev value
    """
    if len(lst) < 2:
        return None
    mean = sum(lst)/float(len(lst))
    return math.sqrt(sum([(x-mean)**2 for x in lst])/float(len(lst)-1.0))

def running_stddev(lst, radius):
    """
    Running stddev. The running stddev for an element n is the stddev of element
    n and radius elements before and after (Window size = 2 * radius + 1).
    For the first (and last) n elements, the number of elements before (after)
    is of course shorter then n.

    Parameters:
        lst    list of values
        radius the radius (how many elements left and right on an element n
               should be included).
    Returns:
        list with running stddev.

    A double-ended queue (deque; with fast append and pop on both sides) is used
    to keep track of the current window.

    Welford's a algorithm (streaming calculation) is used to improve
    performance (see class StddevWeford).
    """
    result = []
    # Initial window: n elements right of the value plus the value itself (=n+1
    # values).
    # We preload the welford object with n values. in the first iteration + 1
    # value is added
    welford = StddevWelford()
    for item in lst[0:radius]:
        welford.push(item)

    for item in lst[radius:]:
        # Iterate through elements from radius to the end
        # Item is the rightmost element to be appended to the window:
        welford.push(item)
        # Delete one element if the window size is exceeded:
        if len(welford.data) > 2 * radius + 1:
            welford.pop()
        # Calculate and append the stddev value:
        # (we know the window contains no None values, use
        # _stddev_fast directly)
        result.append(welford.stddev)
    # Process the final elements (no new values, but pop values from the
    # left of the window.
    # Usually this is radius iterations, but specail case when list size <
    # rarius.
    for i in range(len(lst) - len(result)):
        # with list much longer then radius, we would just unconditionally pop
        # one element per loop. But when list length < 2*r+1 (i.e. full window
        # size is never reached), we pop elements too early for the last r
        # elements. Thus we depend on i:.
        # Example: [1., 2., 3., 4.], radius = 2
        if len(welford.data) > 2 * radius - i:
            welford.pop()
        # Calculate and append the stddev value:
        # (we know the window is sorted and contains no None values, use
        # _stddev_fast directly)
        result.append(welford.stddev)
    return result

class StddevWelford(object):
    """
    Implementation of Welford's algorithm for calculating the stddev.
    This is a numerically quite stable version according to
    https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm

    Added the pop method for removing elements: pop is implemented trivially
    as the opposite of push.

    The Welford algorithm is likely slower for calculating a single stddev.
    But it can calculate streaming data (no need to store all data at once), and
    it has a husge performance advantage for calculating a running stddev (see
    methon running_stddev above).
    """
    def __init__(self):
        """
        Set up the object.
        """
        # sum of the squares of the diffs from mean:
        self.sum_squ_diff_mean = 0.0
        self.data = deque([])
        self.count = 0
        self.mean = 0.0

    def push(self, value):
        """
        Append a value to the end of the list of data.
        """
        self.data.append(value)
        if value is not None:
            self.count += 1
            delta = value - self.mean
            self.mean += delta / self.count
            self.sum_squ_diff_mean += delta * (value - self.mean)

    def pop(self):
        """
        Pop one value from the beginning of the list of data.
        """
        if len(self.data) == 1:
            self.__init__()
        value = self.data.popleft()
        if value is not None:
            self.count -= 1
            delta = value - self.mean
            self.mean -= delta / self.count
            self.sum_squ_diff_mean -= delta * (value - self.mean)

    @property
    def stddev(self):
        """
        Get the resulting stddev.
        """
        if self.count < 2:
            return None
        return math.sqrt(self.sum_squ_diff_mean/float(self.count-1.0))


def percentile(lst, perc, is_sorted=False, contains_none=True):
    """
    A raw and fast version of calculating a percentile of a list of values.
    Parameters:
        lst         list of values. MUST BE already sorted.
                    either float or Decimal, but NOT MIXED!
        perc        percentile to be calculated float 0.0 - 100.0.
        is_sorted   bool, speed up the calculation if the list is already sorted
        contains_none
                    bool, speed up the calculation if the list does not contain
                    Nones
    Returns:
        the percentile of the values
    """
    if contains_none:
        lst = [x for x in lst if x is not None]
    if not is_sorted:
        lst = sorted(lst)
    return _percentile_fast(lst, perc)

def _percentile_fast(lst, perc):
    """
    A raw and fast version of calculating a percentile of a list of values.
    !!! Fast version, list must be sorted!
    For unsorted lists use stddev()
    Parameters:
        lst    list of values. MUST BE already sorted.
               either float or Decimal, but NOT MIXED!
        perc   percentile to be calculated float 0.0 - 100.0.
    Returns:
        the percentile of the values
    """
    if len(lst) == 0:  # pylint: disable=C1801
        # C1801: Do not use `len(SEQUENCE)` to determine if a sequence is  empty
        # --> want to raise TypeError on not sequence objects, e.g. False, None)
        return None

    # float index for the percentile value:
    index = (len(lst)-1) * perc / 100.0
    floor = math.floor(index)
    ceil = math.ceil(index)

    if floor == ceil:
        # index is integer
        return lst[int(index)]
    # else, return the weighed average of the two elements (weighed by distance)
    if isinstance(lst[0], Decimal):
        # special case for Decimal: can only be multiplied by a Decimal
        return lst[int(floor)] * Decimal(ceil-index) + \
               lst[int(ceil)] * Decimal(index-floor)
    return lst[int(floor)] * (ceil-index) + \
           lst[int(ceil)] * (index-floor)



"""
EBAS submission templates and template checks: Ozone Template
"""

from .base import EbasTemplateBase, NoTemplate


class EbasTemplateOzone(EbasTemplateBase):
    """
    Ebas template for ozone
    """

    TEMPLATE_NAME = 'Ozone'

    def match(self):
        """
        See if this template matches a file.
        """
        # Ozone Template exists only for Lev 2
        # Contains only ozone variables
        comp_names = [x.metadata.comp_name for x in self.file.variables]
        if not any([x == 'ozone' for x in comp_names]):
            raise NoTemplate()
        if not all([x == 'ozone' for x in comp_names]):
            self.warning(
                'Ozone template: no match because of variobles: {}'.format(
                    ', '.join([x.metadata.comp_name
                               for x in self.file.variables
                               if x != 'ozone'])))
            raise NoTemplate()
        if self.file.metadata.datalevel != '2':
            self.warning(
                'Ozone template: no match because of data level {}'.format(
                    self.file.metadata.datalevel))
            raise NoTemplate()

    def special_checks(self):
        """
        Special checks which cannot be done in the defailt checks.
        """
        for vnum, var in enumerate(self.file.variables):
            if var.metadata.comp_name == 'ozone':
                file_vnum, _, file_metalnum = self.file.msgref_vnum_lnum(vnum)
                if not self.file.get_meta_for_var(vnum, 'vol_std_temp'):
                    self.error(
                        ("Variable {}: 'Volume std. temperature' metadata "
                         "missing").format(file_vnum), lnum=file_metalnum)
                if not self.file.get_meta_for_var(vnum, 'vol_std_pressure'):
                    self.error(
                        ("Variable {}: 'Volume std. pressure' metadata "
                         "missing").format(file_vnum), lnum=file_metalnum)
                if not self.file.get_meta_for_var(vnum, 'abs_cross_section'):
                    self.error(
                        ("Variable {}: 'Absorption cross section' metadata "
                         "missing").format(file_vnum), lnum=file_metalnum)

                # cross check standard method and absorption cross section:
                # 2023-06-07: AGH maybe this is not true, double check GAW 209!
                # from ebas.io.file.basefile.base import isvalid
                # acs = self.file.get_meta_for_var(vnum, 'abs_cross_section')
                # sme = self.file.get_meta_for_var(vnum, 'std_method')
                # if (sme == 'SOP=GAW_209(2013)' and isvalid(acs) and 
                #     acs != 'Hearn, 1961'):
                #     self.error(
                #         ("Variable {}: Standard method 'SOP=GAW_209(2013)' "
                #          "does not allow absorption cross section '{}' (only "
                #          "'Hearn, 1961')").format(
                #              file_vnum, acs), lnum=file_metalnum)



#!/Library/Frameworks/Python.framework/Versions/2.7/bin/python
import sys
from . import ebas
from . import gml
import datetime
res = ebas.ResultSetType()
res.set_TimeDimension(ebas.TimeDimensionType(
   [datetime.datetime(2011,1,1), datetime.datetime(2011, 1, 2), datetime.datetime(2011,1,3), datetime.datetime(2011, 1, 4)],
   [datetime.datetime(2011,1,2), datetime.datetime(2011, 1, 3), datetime.datetime(2011,1,4), datetime.datetime(2011, 1, 5)]))


position=ebas.Position(gml.PointType(id='pos1', srsName='urn:ogc:def:crs:EPSG:6.6:4979', srsDimension=3, pos=gml.DirectPositionType(valueOf_='1 2 3')))
res.add_DataSet(ebas.DataSetType(123456, 'NO0001R', 'Birkenes', 'filterpack', 'sulfate', Matrix='pm10', Unit='mg/m3', Statistics='arithmetic mean', ResolutionCode='1d', DataLevel='2', Position=position, HeightAGL=ebas.HeightAGL('m', 11), Values=(1,2,3,4,5)))
result = ebas.Result()
result.add_ResultSet(res)

sys.stdout.write('<?xml version="1.0" ?>\n')
result.export(sys.stdout, 0, namespacedef_='xmlns:ebas="http://ebas.nilu.no/EBAS" xmlns:gml="http://www.opengis.net/gml/3.2"')



"""
$Id: indexdb_internal.py 2266 2019-04-10 07:42:19Z pe $

DB layer for fileindex db.
The fileindex IndexDBInternal is a sqlite3 DB used internally in ebas.
"""
from textwrap import dedent
from databases.sqlite import SqliteDb


class IndexDbInternal(SqliteDb):
    """
    EBAS fileindex DB I/O class.
    """

    def __init__(self, filename, timestamp):
        """
        Initialize DB connection. Open DB, create tables.
        Parameters:
            filename   filename for the database. If prepended with '+', ignore
                       errors while creating tables.
        """
        super(IndexDbInternal, self).__init__(filename)
        self.timestamp = timestamp

    def insert_file_obj(self, ebasfile):
        """
        Main trigger method: perform all inserts reated to the data file.
        Parameters:
            ebasfile   ebas.io.file.EbasFile object
        Returns:
            None
        """
        file_dict = {
             'filename': ebasfile.metadata.filename,
             'first_start': ebasfile.sample_times[0][0],
             'first_end': ebasfile.sample_times[0][1],
             'last_start': ebasfile.sample_times[-1][0],
             'last_end': ebasfile.sample_times[-1][1],
        }
        self.insert_file(file_dict)

        var_meta = ('setkey', 'filename')
        for i in range(len(ebasfile.variables)):
            var_dict = {key: ebasfile.get_meta_for_var(i, key)
                        for key in var_meta}
            self.insert_variable(var_dict)

    def create_tables(self):
        """
        Create all necessary tables.
        Parameters:
            None
        Returns:
            None
        Raises:
            IndexDbExists  when the tables already exist
        """
        self.create_table_update()
        self.create_table_file()
        self.create_table_variable()
        self.dbh.commit()
        self.logger.debug("tables created")

    def create_table_update(self):
        """
        Create file table.
        """
        self.dbh.execute(dedent("""\
            create table update_log
            (
               update_ts timestamp primary key)
            """))

    def create_table_file(self):
        """
        Create file table.
        """
        self.dbh.execute(dedent("""\
            create table file
            (
               filename TEXT not NULL primary key,
               first_start TEXT not NULL,
               first_end TEXT not NULL,
               last_start TEXT not NULL,
               last_end TEXT not NULL)
            """))

    def create_table_variable(self):
        """
        Create variable table.
        """
        self.dbh.execute(dedent("""\
            create table variable
            (
               setkey INTEGER not NULL,
               filename TEXT not NULL)
            """))

    def insert_update(self, update_dict):
        """
        Insert file metadata into the IndexDbInternal.
        """
        self.dbh.execute(dedent("""\
            insert into update_log values
               (:update_ts)
            """), update_dict)
        self.dbh.commit()

    def get_last_update(self):
        """
        Get the last update date from the index database.
        """
        curs = self.dbh.execute(
            "select max(update_ts) as 'update_ts [timestamp]' from update_log")
        # !!!
        # as 'update_ts [timestamp]'
        # is necessary to trigger conversion from timestamp to datetime.dateime
        return curs.fetchall()[0][0]

    def expand_reexport(self, setkeys):
        """
        Expands the set of data to be re-exported.
        Parameters:
            setkeys   set of setkeys to be re-exported (will be extended)
        Returns:
            tuple (set of filenames to be deleted,
                   set of datasets to be exported)
        """
        filenames = set()
        curs = self.dbh.cursor()
        new_setkeys = setkeys.copy()
        self.logger.info("%d setkeys for reexport: %s",
                          len(new_setkeys), new_setkeys)
        while new_setkeys:
            # select all files where the setkeys are in:
            curs = self.dbh.execute(dedent("""\
                select distinct filename
                from variable
                where setkey in ({})
                """.format(','.join('?'*len(new_setkeys)))),
                list(new_setkeys))
            new_filenames =  [x[0] for x in curs.fetchall()]
            filenames = filenames.union(new_filenames)
            self.logger.info("%d files affected: %s",
                              len(new_filenames), new_filenames)
            # ... and mark the new_setkeys as deleted
            self.dbh.execute(dedent("""\
                delete from variable
                where setkey in ({})
                """.format(','.join('?'*len(new_setkeys)))),
                tuple(new_setkeys))

            # and add the new setkeys:
            setkeys = setkeys.union(new_setkeys)

            # now select all setkeys which were in the same files
            curs = self.dbh.execute(dedent("""\
                select distinct setkey
                from variable
                where filename in ({})
                """.format(','.join('?'*len(new_filenames)))),
                tuple(new_filenames))
            new_setkeys = [x[0] for x in curs.fetchall()]
            self.logger.info("%d additional setkeys in the same files: %s",
                              len(new_setkeys), new_setkeys)
            # mark the files as deleted:
            if new_filenames:
                self.dbh.execute(dedent("""\
                    delete from file
                    where filename in ({})
                    """.format(','.join('?'*len(new_filenames)))),
                    tuple(new_filenames))
        return filenames, setkeys

    def insert_file(self, file_dict):
        """
        Insert file metadata into the IndexDbInternal.
        """
        file_dict = file_dict.copy()
        self.dbh.execute(dedent("""\
            insert into file
               (filename, first_start, first_end, last_start, last_end)
            values
               (:filename, :first_start, :first_end, :last_start, :last_end)
            """), file_dict)
        self.dbh.commit()

    def select_oldfiles(self, timestamp):
        """
        Selects all files which do not have any data after timestamp.
        Parameters:
            timestamp      timestamp for selecting old files
        Returns:
            iterator of filenames
        """
        curs = self.dbh.execute(
            "select filename from file where last_end < ?",
            (timestamp,))
        for res in curs.fetchall():
           yield res[0]

    def delete_oldfile(self, filename):
        """
        Delete old file (cascade variables) from the index db.
        Parameters:
            filename    filename of the file
        """
        self.dbh.execute(dedent("""\
            delete from variable
            where filename=?"""),
            (filename,))
        self.dbh.execute(dedent("""\
            delete from file
            where filename=?"""),
            (filename,))

    def insert_variable(self, var_dict):
        """
        Insert metadata for one variable of one file into the IndexDbInternal.
        """
        var_dict = var_dict.copy()
        self.dbh.execute(dedent("""\
            insert into variable
               (setkey, filename)
            values
               (:setkey, :filename)
            """), var_dict)
        self.dbh.commit()



"""
$Id: sx.py 2743 2021-11-23 15:48:15Z pe $
"""

from __future__ import absolute_import
# otherwise re is imported relative as masterdata.re
import re
import six
from .offline_masterdata import SXOfflineMasterData
from .base import EbasMasterBase
from .sz import EbasMasterSZ

class EbasMasterSX(EbasMasterBase, SXOfflineMasterData):
    """
    Domain Class for station auxiliary masterdata.
    Objects of this class do not represent entities, this class provides class
    methods handling caches masterdata.
    Stations master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for SX (station auxiliary masterdata)
    # Those are fallback values, will be read from database as soon as possible.
    SXOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read st masterdata if dbh is provided.
        """
        SXOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)



"""
ebas/io/base.py
$Id: consolidate.py 2731 2021-11-11 00:04:05Z pe $

Base classes for (file) I/O.

"""

import datetime
from nilutility.datatypes import Histogram
from .base import EbasFileBase
from ..base import FLAGS_ONE_OR_ALL, FLAGS_COMPRESS, FLAGS_ALL, FLAGS_NONE, \
    FLAGS_AS_IS, \
    SUPPRESS_SORT_VARIABLES, SUPPRESS_METADATA_OCCURRENCE




class EbasFileConsolidate(EbasFileBase):  # pylint: disable=W0223
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    Partial class for EbasFile (Base class for file objects).
    This part handles file consolidation (standardized reformatting before
    output)
    """

    def consolidate_metadata(self, flags=FLAGS_ONE_OR_ALL, suppress=0):
        """
        Optimize the metadata structure of the object.
        - Order variables
        - Move attibutes between global (NNCOM) and variable specific (VNAME)
        - Choose the maximum revdate of all variables and set as file revdate
        - Eliminate redundant flag columns
        - Generate the filename for metadata (according to all other metadata)
        Parameters:
            flags         flag column options: FLAGS_ONE_OR_ALL (default),
                          FLAGS_COMPRESS, FLAGS_ALL, FLAGS_NONE
            suppress      suppress selected consolidations (bitfield):
                              SUPPRESS_SORT_VARIABLES
                              SUPPRESS_METADATA_OCCURRENCE
        Returns:
            None
        """
        # set reference date if not set:
        if not self.metadata.reference_date and self.sample_times:
            self.metadata.reference_date = \
                datetime.datetime(self.sample_times[0][1].year, 1, 1)

        self._set_method_analytical()

        # strict_gloabl and SUPPRESS_METADATA_OCCURRENCE is a contradiction
        if self.strict_global and suppress & SUPPRESS_METADATA_OCCURRENCE:
            suppress &= ~SUPPRESS_METADATA_OCCURRENCE
            self.logger.warning(
                "SUPPRESS_METADATA_OCCURRENCE is overruled by strict_global")
        if not suppress & SUPPRESS_METADATA_OCCURRENCE:
            for key in self.metadata.keys():
                if key in ('title', 'revdate', 'characteristics', 'qa'):
                    # title: is a per variable metadata element
                    # revdate: special behaviour, see _consolidate_revdate()
                    # characteristics: maybe implement special behaviour later
                    # qa: special behaviour, see _consolidate_qa()
                    continue
                self._consolidate_metadata_occ(key)
        self._consolidate_revdate()
        if not suppress & SUPPRESS_METADATA_OCCURRENCE:
            self._consolidate_qa()
        if not suppress & SUPPRESS_SORT_VARIABLES:
            self._sort_variables()
        # set proper revdate
        self._consolidate_flagcolumns(flags)
        self._consolidate_title()
        self.gen_filename()

    def _sort_variables(self):
        """
        Resort variables according to standard.
        Sort order:
            A) categories
                1) auxiliary data
                2) comp = precipitation amount
                3) others
            B) instrument name
            C) sort order for special parameters (mainly aux data)
                start_time, end_time, number of size bins,
                latitude, longitude, altitude, other position
            D) rest is sorted by comp name, characteristics, statistics and
               variable metadata:
                comp name
                characteristics
                statistics
                local metadata
        """
        scorelist = []
        for i in range(len(self.variables)):
            score = []
            matrix = self.get_meta_for_var(i, 'matrix')
            comp = self.get_meta_for_var(i, 'comp_name')

            # A: categories
            if self.is_auxiliary(i):
                score.append(1)
            elif comp and comp.startswith('precipitation_amount'):
                score.append(2)
            else:
                score.append(100)

            # B: instrument name
            score.append(self.get_meta_for_var(i, 'instr_name'))

            # C: sort order for special parameters (mainly auxiliary variables)
            if comp == 'start_time':
                # skipped on import, but when creating file it still needs to
                # get the right order.
                score.append(1)
            elif comp == 'end_time':
                # skipped on import, but when creating file it still needs to
                # get the right order.
                score.append(2)
            elif comp == 'number of size bins':
                # skipped on import, but when creating file it still needs to
                # get the right order.
                score.append(3)
            elif comp == 'latitude':
                score.append(4)
            elif comp == 'longitude':
                score.append(5)
            elif comp == 'altitude':
                score.append(6)
            elif matrix == 'position':
                # position, but not lat, lon, alt?
                score.append(7)
            elif matrix == 'instrument':
                score.append(8)
            else:
                score.append(100)

            # D:
            score.append(comp)
            # append order number as well as tag, value and unit as a tuple
            # that way, the most important ct_type (by order number) gets
            # prioritized.
            characteristics = []
            if 'characteristics' in self.variables[i].metadata and \
                    self.variables[i].metadata.characteristics is not None:
                for ordered in self.variables[i].metadata.characteristics.\
                                                                sorted_order():
                    characteristics.append((ordered[0], ordered[1].tuple()))
            score.append(tuple(characteristics))
            score.append(self.get_meta_for_var(i, 'statistics'))
            metadata = []
            for meta in self.internal.ebasmetadata.metadata_list_tag_value(
                    self, i):
                metadata.append((meta[2]['sortorder'], meta[0], meta[1]))
            score.append(tuple(metadata))
            score.append(i)  # last element is index
            scorelist.append(score)
        scorelist.sort()
        new_variables = []
        for sort_var in scorelist:
            new_variables.append(self.variables[sort_var[-1]])
        self.variables = new_variables

    def _consolidate_revdate(self):
        """
        Set the maximum revdate of all variables as the default revdate for
        the file. This is the expected behaviour for EBAS NASA-Ames.
        Parameters:
            None
        Returns:
            None
        """
        # main revdate is maximum of variables revdate's
        lst = [v.metadata.revdate for v in self.variables \
               if 'revdate' in v.metadata] + \
              [self.metadata.revdate for v in self.variables \
               if 'revdate' not in v.metadata]
        lst = [x for x in lst if x]
        if lst:
            newvalue = max(lst)
        else:
            newvalue = None
        # set new value, change all var metadata if necessary
        # this should be done anyway, there could be some var metadata
        # equal default
        for var in self.variables:
            if 'revdate' in var.metadata and \
               var.metadata['revdate'] == newvalue:
                # var metadata is equal to new default
                del var.metadata['revdate']
            elif 'revdate' not in var.metadata and \
               self.metadata.revdate != newvalue:
                # Set old default to var metadata.
                # (this would also set it to None, in case the old default
                # was None and the new default is different)
                var.metadata.revdate = self.metadata.revdate
            # else: leave as is
        # set the new default:
        self.metadata.revdate = newvalue

    def _consolidate_metadata_occ(self, key):
        """
        Move attibutes between global (NNCOM) and variable specific (VNAME).
        Parameters:
            key           metadata key name
        Returns:
            None
        """
        # hist will be a histogram of used values
        hist = Histogram()
        hist_noaux = Histogram()  # histogram for auxiliary variables excluded
        cnt_noaux = 0
        for vnum, var in enumerate(self.variables):
            if key in var.metadata and \
               var.metadata[key] != self.metadata[key]:
                hist.increment(var.metadata[key])
                if not self.is_auxiliary(vnum):
                    hist_noaux.increment(var.metadata[key])
                    cnt_noaux += 1
            else:
                hist.increment(self.metadata[key])
                if not self.is_auxiliary(vnum):
                    hist_noaux.increment(self.metadata[key])
                    cnt_noaux += 1

        # check all vars have the same rescode
        # if it differs, it's probably a metadata error in the DB, throw warning
        (maxocc, values) = hist.get_max()
        if key == 'resolution' and maxocc != len(self.variables):
            self.warning('Different resolution codes in one file')
            # see below, we need to provide a value anyway!

        # hist_noaux should only be used, if there is at least one element:
        if not self.strict_global and hist_noaux.get_max()[0] > 0:
            hist = hist_noaux
            cnt = cnt_noaux
        else:
            cnt = len(self.variables)
        # now decide if there should be a "default" value (main metadata)
        # current criteria: must be used by more than half of the variables
        (maxocc, values) = hist.get_max()
        if maxocc == cnt:
            # this applies to strict_global == True
            newvalue = values[0]
        elif not self.strict_global and maxocc > cnt / 2:
            # more than halv the variables have the same metadata value:
            newvalue = values[0]
        elif key == 'resolution':
            # special case: resolution: __must__ by definition be file global
            # see above, this should never happen, but we need a value anyway
            newvalue = values[0]
        else:
            newvalue = None
        # set new value, change all var metadata if necessary
        # this should be done anyway, there could be some var metadata
        # equal default
        for var in self.variables:
            if key in var.metadata and \
               var.metadata[key] == newvalue:
                # var metadata is equal to new default
                if key not in ('comp_name', 'unit'):
                    # comp name and unit must occur in every variable)
                    del var.metadata[key]
            elif key not in var.metadata and self.metadata[key] != newvalue:
                # Set old default to var metadata.
                # (this would also set it to None, in case the old default
                # was None and the new default is different)
                var.metadata[key] = self.metadata[key]
            # else: leave as is
        # set the new default:
        self.metadata[key] = newvalue

    def _consolidate_qa(self):
        """
        Move QA attibutes between global (NNCOM) and variable specific (VNAME).
        Parameters:
            None
        Returns:
            None
        """
        for elem in self.internal.ebasmetadata.metadata:
            key = elem['key']
            if 'QA_block' in elem and elem['QA_block']:
                for qanum in self.get_qa_numbers():
                    hist = Histogram()
                    for vnum in range(len(self.variables)):
                        for qa_ in self.get_qa_for_var(vnum):
                            if qa_['qa_number'] == qanum and key in qa_:
                                hist.increment(qa_[key])

                    # now decide if there should be a "default" value (main
                    # metadata) current criteria: must be used by more than half
                    # of the variables
                    (maxocc, values) = hist.get_max()
                    if maxocc > len(self.variables) / 2:
                        newvalue = values[0]
                    else:
                        newvalue = None

                    # set new value, change all var metadata if necessary
                    # this should be done anyway, there could be some var
                    # metadata equal default
                    for vnum in range(len(self.variables)):
                        qa_ = self.get_qa_by_number(qanum, vnum)
                        if qa_ and key in qa_ and qa_[key] == newvalue:
                            # var metadata is equal to new default
                            self.unset_qa(key, qanum, vnum)
                        elif not qa_ or key not in qa_:
                            main_qa = self.get_qa_by_number(qanum, None)
                            if main_qa and key in main_qa and \
                               main_qa[key] != newvalue:
                                # Set old default to var metadata.
                                # (this would also set it to None, in case the
                                # old default was None and the new default is
                                # different)
                                self.set_qa(key, main_qa[key], qanum, vnum)
                        # else: leave as is
                    # set the new default:
                    self.set_qa(key, newvalue, qanum, None)


    def _consolidate_flagcolumns(self, flags=FLAGS_ONE_OR_ALL):
        """
        Eliminates redundantflag columns.
        Parameters:
            flags         flag column options: FLAGS_ONE_OR_ALL (default),
                          FLAGS_COMPRESS, FLAGS_ALL, FLAGS_NONE
        """
        _cleanflags = lambda flags: [sorted([y for y in x if y != 999])
                                     for x in flags]

        latestflags = None
        manual_latest = None
        for i in reversed(range(len(self.variables))):
            if flags == FLAGS_NONE:
                self.variables[i].flagcol = False
                # clean values (set invalid to missing, set 781 to 1/2 value)
                # is done only on export - so the IO object remains
                # fully valid!
                # see: nasa_ames_write:set_data()
            elif flags == FLAGS_ALL:
                self.variables[i].flagcol = True
            # for FLAGS_ONE_OR_ALL, FLAGS_COMPRESS and FLAGS_AS_IS:
            # The flags need to be modified in a special case on output:
            # When multiple variables share a flag column and the flag column
            # has a flag 999 in a sample where one of the variables has a valid
            # value: the 999 needs to be dropped in this case.
            # But we do not change the flag sequence of the valid I/O object
            # here, this is been taken care of on flag output (for nasa_ames in
            # the set_data / _gen_flags methods in ebas/io/nasa_ames/write.py
            elif flags == FLAGS_ONE_OR_ALL:
                if latestflags is None:
                    latestflags = _cleanflags(self.variables[i].flags)
                    self.variables[i].flagcol = True
                elif latestflags == _cleanflags(self.variables[i].flags):
                    self.variables[i].flagcol = False
                else:
                    # flaglist differs, recursive call with FLAGS_ALL and return
                    self._consolidate_flagcolumns(flags=FLAGS_ALL)
                    return
            elif flags == FLAGS_COMPRESS:
                if latestflags == _cleanflags(self.variables[i].flags):
                    self.variables[i].flagcol = False
                else:
                    latestflags = _cleanflags(self.variables[i].flags)
                    self.variables[i].flagcol = True
            elif flags == FLAGS_AS_IS:
                # The warnings here could be errors.
                # Not sure, if someone uses FLAGS_AS_IS and does not stick to
                # the flagging rules, should it fail?
                # For now, we are graceful and do the best we can to fix it.
                file_vnum, _, _ = self.msgref_vnum_lnum(i)
                if i == len(self.variables) - 1:
                    if not self.variables[i].flagcol:
                        self.warning(
                            "Variable {}: last variable must have a flag "
                            "column".format(file_vnum))
                        self.variables[i].flagcol = True
                    manual_latest = i
                    latestflags = _cleanflags(self.variables[i].flags)
                elif not self.variables[i].flagcol and \
                        latestflags != _cleanflags(self.variables[i].flags):
                    file_vnum_flagcol, _, _ = self.msgref_vnum_lnum(
                        manual_latest)
                    self.warning(
                        "Variable {}: flag sequence is not consistent with "
                        "flag sequence of variable {}. Adding flag column for "
                        "variable {}.".format(
                            file_vnum, file_vnum_flagcol, file_vnum))
                    self.variables[i].flagcol = True
                    manual_latest = i
                    latestflags = _cleanflags(self.variables[i].flags)
                else:
                    # self.variables[i].flagcol == True
                    manual_latest = i
                    latestflags = _cleanflags(self.variables[i].flags)

    def _consolidate_title(self):
        """
        Generates a title string if none is set already.
        """
        for i in range(len(self.variables)):
            if 'title' not in self.variables[i].metadata or \
               self.variables[i].metadata.title is None:
                self.variables[i].metadata.title = \
                    self.variables[i].metadata.comp_name \
                        if "comp_name" in self.variables[i].metadata else ''

    def _set_method_analytical(self):
        """
        Sets the method ref if not set and analytical metadata instrument are
        given.
        Parameters:
            None
        Returns:
            None
        """
        def _get_ana_meta(vnum, pure=True):
            """
            Helper for getting metadata for a specific variable number.
            Parameters:
                vnum     variable index (None = NNCOM metadata)
                pure     return pure vname specifications for a variable
                         (might be different from the effective metadata set)
            Returns
                metadata dict for variable
            """
            tags = (
                'method', 'ext_lab', 'ana_technique', 'ana_lab_code',
                'ana_instr_name', 'ana_instr_manufacturer', 'ana_instr_model',
                'ana_instr_serialno')
            if vnum != None and not pure:
                return {tag: self.get_meta_for_var(vnum, tag) for tag in tags}

            metadata = self.variables[vnum].metadata if vnum != None \
                           else self.metadata
            return {tag: metadata[tag] for tag in tags if tag in metadata}

        # check analytical metadata consistency in NCOM:
        nnmeta = _get_ana_meta(None)

        for vnum in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(vnum)
            meta = _get_ana_meta(vnum)
            if meta:
                # something has been defined on variable level: check again for
                # this variable
                meta = _get_ana_meta(vnum, False)
                if ('method' not in self.variables[vnum].metadata or \
                   not self.variables[vnum].metadata['method']) and \
                   meta['ana_lab_code']  and meta['ana_instr_name']:
                    self.variables[vnum].metadata['method'] = \
                        meta['ana_lab_code'] + '_' + meta['ana_instr_name']
                    self.warning(
                        "Variable {}: automatically generating 'Method ref' "
                        "from analytical metadata: {}".format(
                            file_vnum, self.variables[vnum].metadata['method']),
                        lnum=file_metalnum)

                elif meta['method'] is None:
                    self.error(
                        "Variable {}: Missing metadata 'Method ref'".format(
                            file_vnum), lnum=file_metalnum)

        if not nnmeta['method'] and nnmeta['ana_lab_code']  and \
           nnmeta['ana_instr_name']:
            self.metadata['method'] = \
                nnmeta['ana_lab_code'] + '_' + nnmeta['ana_instr_name']
            self.warning(
                "automatically generating 'Method ref' from analytical "
                "metadata: {}".format(self.metadata['method']))



"""
EBAS I/O file module
"""
from six import string_types
from collections import OrderedDict
from .base import FLAGS_ONE_OR_ALL, FLAGS_COMPRESS, FLAGS_ALL, FLAGS_NONE, \
    FLAGS_AS_IS, \
    EBAS_IOFORMAT_NASA_AMES, EBAS_IOFORMAT_CSV, EBAS_IOFORMAT_XML, \
    EBAS_IOFORMAT_NETCDF1, EBAS_IOFORMAT_NETCDF, EBAS_IOFORMAT_OPENDAP, \
    EBAS_IOSTYLE_SINGLECOLUMN, EBAS_IOSTYLE_MULTICOLUMN, \
    EBAS_IOSTYLE_MULTICOLUMN_PRECIP, EBAS_IOSTYLE_KEEP, \
    EBAS_IOMETADATA_OPTION_SETKEY, EBAS_IOMETADATA_OPTION_ALL, \
    SUPPRESS_SORT_VARIABLES, SUPPRESS_METADATA_OCCURRENCE

class BaseEbasIOOption(int):
    """
    Base class for options (commandline):
    """
    options = None
    roptions = None

    def __new__(cls, *args):
        if len(args) == 1 and isinstance(args[0], string_types):
            try:
                obj = int.__new__(cls, cls.options[args[0]])
            except KeyError:
                raise ValueError("invalid literal for {}: '{}'".format(
                    cls.__name__, args[0]))
        else:
            obj = int.__new__(cls, *args)
        return obj
    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               self.__class__.roptions[self])
    def __str__(self):
        return '{}'.format(self.__class__.roptions[self])

    @classmethod
    def legal_options(cls):
        """
        Returns a list of option names.
        """
        return [cls(x) for x in cls.options.keys()]


class EbasIOFlagOption(BaseEbasIOOption):
    """
    Flag options for NILU Nasa Ames files:
    """
    options = OrderedDict([
        ('one-or-all', FLAGS_ONE_OR_ALL),
        ('compress', FLAGS_COMPRESS),
        ('all', FLAGS_ALL),
        ('none', FLAGS_NONE),
    ])
    roptions = OrderedDict([(x[1], x[0]) for x in options.items()])

class EbasIOFormatOption(BaseEbasIOOption):
    """
    Format options for EBAS files:
    """
    options = OrderedDict([
        ('NasaAmes', EBAS_IOFORMAT_NASA_AMES),
        ('CSV', EBAS_IOFORMAT_CSV),
        ('XML', EBAS_IOFORMAT_XML),
        ('NetCDF1', EBAS_IOFORMAT_NETCDF1),
        ('NetCDF', EBAS_IOFORMAT_NETCDF),
        ('OPeNDAP', EBAS_IOFORMAT_OPENDAP),
    ])
    roptions = OrderedDict([(x[1], x[0]) for x in options.items()])



"""
EBAS IO File base definistions
"""

EBAS_IOFORMAT_NASA_AMES = 0
EBAS_IOFORMAT_CSV = 1
EBAS_IOFORMAT_XML = 2
EBAS_IOFORMAT_NETCDF1 = 3
EBAS_IOFORMAT_NETCDF = 4
EBAS_IOFORMAT_OPENDAP = 5

EBAS_IOSTYLE_SINGLECOLUMN = 0
EBAS_IOSTYLE_MULTICOLUMN = 1
EBAS_IOSTYLE_MULTICOLUMN_PRECIP = 2
EBAS_IOSTYLE_KEEP = 3

# bitwise options:
EBAS_IOMETADATA_OPTION_SETKEY = 1
EBAS_IOMETADATA_OPTION_ALL = \
    EBAS_IOMETADATA_OPTION_SETKEY

FLAGS_ONE_OR_ALL = 0
FLAGS_COMPRESS = 1
FLAGS_ALL = 2
FLAGS_NONE = 3
FLAGS_AS_IS = 4

SUPPRESS_SORT_VARIABLES = 1
SUPPRESS_METADATA_OCCURRENCE = 2

